@@ -11,7 +11,7 @@
 
 ********************************************************************/
 /* ***** BEGIN LICENSE BLOCK *****
- * Source last modified: $Id: rvorbis.cpp,v 1.8.2.4 2004/07/09 01:58:21 hubbe Exp $
+ * Source last modified: $Id: rvorbis.cpp,v 1.8.2.6 2004/11/24 18:29:53 acolwell Exp $
  * 
  * Portions Copyright (c) 1995-2004 RealNetworks, Inc. All Rights Reserved.
  * 
@@ -77,15 +77,27 @@
 #include "hxplugn.h"
 #include "hxrendr.h"
 #include "hxausvc.h"
+#include "hxmon.h"
 #include "hxtick.h" // IsTimeLess(), IsTimeGreaterOrEqual()
+#include "pckunpck.h"
 #include "rvorbis.h"
 
 #include "ogg_depack.h" // OggDepacketizer
 
+#include "hxassert.h"
+#include "debug.h" // DPRINTF()
+#define D_VORBIS 0 //0x2000000
+
 #define NO_TIME_SET 0xffffffff
 
+const UINT32 InitialPushdownDuration = 500;
+
 STDAPI RMACreateInstance(IUnknown **ppRendererObj)
 {
+#ifdef _DEBUG
+    debug_level() |= D_VORBIS;
+#endif /* _DEBUG */
+
     *ppRendererObj = (IUnknown *)(IHXPlugin *)new CVorbisRenderer();
     if (*ppRendererObj != NULL) {
         (*ppRendererObj)->AddRef();
@@ -104,6 +116,15 @@ const VorbisDepackInfo CVorbisRenderer::zm_depackInfo[] = {
     {"application/ogg", &OggDepacketizer::Create}
 };
 
+const VorbisTagMapping CVorbisRenderer::zm_tagInfo[] = 
+{
+    {"TITLE", "Title"},
+    {"ARTIST", "Author"},
+    {"AUTHOR", "Author"},
+    {"DESCRIPTION", "Description"},
+    {"COPYRIGHT", "Copyright"}
+};
+
 CVorbisRenderer::CVorbisRenderer(void)
     : m_RefCount(0),
       m_pClassFactory(NULL),
@@ -115,12 +136,13 @@ CVorbisRenderer::CVorbisRenderer(void)
       m_lTimeOffset(0),
       m_VorbisState(VorbisInitialHeader),
       m_totalSamples(0),
+      m_uAudioPushdownStartTS(0),
       m_pContext(NULL),
       m_pPlayer(NULL),
       m_pStream(NULL),
       m_pAudioPlayer(NULL),
-      m_pAudioStream(NULL),
       m_pStreamHeaderObj(NULL),
+      m_pCurrentStream(NULL),
       m_pDepack(NULL),
       m_pStreamMimeTypes(NULL)
 {
@@ -191,7 +213,7 @@ STDMETHODIMP CVorbisRenderer::InitPlugin(IUnknown *pRMACore)
 STDMETHODIMP CVorbisRenderer::StartStream(IHXStream *pStream, 
                                           IHXPlayer *pPlayer)
 {
-    //printf ("CVorbisRenderer::StartStream()\n");
+    DPRINTF(D_VORBIS,("CVorbisRenderer::StartStream()\n"));
 
     m_pStream = pStream;
     m_pPlayer = pPlayer;
@@ -271,7 +293,7 @@ STDMETHODIMP CVorbisRenderer::OnHeader(IHXValues* pStreamHdr)
 
 STDMETHODIMP CVorbisRenderer::OnBegin(UINT32 ulTime)
 {
-    //printf ("OnBegin(%u)\n", ulTime);
+    DPRINTF(D_VORBIS,("OnBegin(%u)\n", ulTime));
 
     return HXR_OK;
 }
@@ -290,8 +312,8 @@ STDMETHODIMP CVorbisRenderer::OnPacket(IHXPacket *pPacketObj,
 {
     HX_RESULT res = HXR_UNEXPECTED;
 
-    //printf("Entering OnPacket() timestamp %u, offset %d\n", 
-    //       pPacketObj->GetTime(), lTimeOffset);
+    DPRINTF(D_VORBIS,("Entering OnPacket() timestamp %u, offset %d\n", 
+                      pPacketObj->GetTime(), lTimeOffset));
 
     if (m_bInSeekMode)
     {
@@ -307,6 +329,13 @@ STDMETHODIMP CVorbisRenderer::OnPacket(IHXPacket *pPacketObj,
 	    // Calculate current position based off the packet timestamp
 	    m_totalSamples = (((ogg_int64_t)pPacketObj->GetTime()) * 
                               m_audioFmt.ulSamplesPerSec) / 1000;
+            
+
+            if (VorbisPlay == m_VorbisState)
+            {
+                m_VorbisState = VorbisWriteInitialPushdown;
+                m_uAudioPushdownStartTS = CurrentTime();
+            }
 
 	    m_bIsGapInStreaming = TRUE;
 	    m_bIsFirstPacket = FALSE;
@@ -326,33 +355,41 @@ STDMETHODIMP CVorbisRenderer::OnPacket(IHXPacket *pPacketObj,
                 if (HXR_OK == err)
                 {
                     switch (m_VorbisState) {
-                    case VorbisInitialHeader:
-                        if (vorbis_synthesis_headerin(&m_vi, &m_vc, pOp) >= 0) 
+                    case VorbisWriteInitialPushdown:
+                        if ((pOp->packet[0] & 0x1) == 0)
                         {
-                            //printf("Got Initial Header\n");
-                            
-                            /* check to see if the audio format is different
-                             * from the default values 
-                             */
-                            if ((m_audioFmt.uChannels != m_vi.channels) ||
-                                (m_audioFmt.ulSamplesPerSec != (UINT32)m_vi.rate))
+                            DPRINTF(D_VORBIS,("Got initial pushdown packet\n"));
+
+                            DecodeAndRender(pOp);
+
+                            if (IsTimeGreaterOrEqual(CurrentTime(),
+                                                     m_uAudioPushdownStartTS + 
+                                                     InitialPushdownDuration))
                             {
-                                /* The parameters are different. 
-                                 * Release our handle on the current stream
-                                 * and create a new audio stream object.
-                                 */
-                                res = InitAudioStream(m_vi.rate, m_vi.channels);
+                                m_VorbisState = VorbisPlay;
                             }
-                            
-                            m_VorbisState = VorbisCommentHeader;
-                            
-                        } 
+                        }
+                        else if (pOp->packet[0] ==  1)
+                        {
+                            gotoInitialHeaderState();
+                            res = handleInitialHeader(pOp);
+                        }
+                        else
+                        {
+                            // This isn't an ident header like
+                            // we expect.
+                            err = HXR_UNEXPECTED;
+                        }
+                        break;
+                    case VorbisInitialHeader:
+                        res = handleInitialHeader(pOp);
                         break;
         
                     case VorbisCommentHeader:
                         if (vorbis_synthesis_headerin(&m_vi, &m_vc, pOp) >= 0)
                         {
-                            //printf("Got Comment Header\n");
+                            DPRINTF(D_VORBIS,("Got Comment Header\n"));
+                            updateTACInfo(&m_vc);
                             m_VorbisState = VorbisCodebookHeader;
                         }
                         break;
@@ -362,9 +399,10 @@ STDMETHODIMP CVorbisRenderer::OnPacket(IHXPacket *pPacketObj,
                         vorbis_synthesis_init(&m_vd, &m_vi);
                         vorbis_block_init(&m_vd, &m_vb);
                         
-                        m_VorbisState = VorbisPlay;
-        
-                        //printf("Got Codebooks\n");                    
+                        m_VorbisState = VorbisWriteInitialPushdown;
+                        m_uAudioPushdownStartTS = CurrentTime();
+
+                        DPRINTF(D_VORBIS,("Got Codebooks\n"));
                         break;
                     };
 
@@ -378,7 +416,8 @@ STDMETHODIMP CVorbisRenderer::OnPacket(IHXPacket *pPacketObj,
             }
         }
     }
-    //printf("Leaving OnPacket() %08x\n", res);
+
+    DPRINTF(D_VORBIS,("Leaving OnPacket() %08x\n", res));
 
     return res;
 }
@@ -413,16 +452,22 @@ STDMETHODIMP CVorbisRenderer::OnDryNotification(UINT32 ulCurrentStreamTime,
 {
     HX_RESULT res = HXR_UNEXPECTED;
 
-    //printf ("OnDryNotification(%u %u)\n", 
-    //        ulCurrentStreamTime, ulMinimumDurationRequired);
+    DPRINTF(D_VORBIS,("OnDryNotification(%u %u)\n", 
+                      ulCurrentStreamTime, ulMinimumDurationRequired));
 
-    if (m_pDepack)
+    if (VorbisPlay != m_VorbisState)
+    {
+        res = HXR_OK;   
+    }
+    else if (m_pDepack)
     {
         BOOL bDone = FALSE;
-        
+
         UINT32 ulEndTime = ulCurrentStreamTime + ulMinimumDurationRequired;
         BOOL bWroteData = FALSE;
         UINT32 ulLastWriteTime = 0;
+        BOOL bAudioParamsChanged = FALSE;
+        BOOL bEOS = FALSE;
         
         while (!bDone)
         {
@@ -431,14 +476,60 @@ STDMETHODIMP CVorbisRenderer::OnDryNotification(UINT32 ulCurrentStreamTime,
             
             if (HXR_OK == err)
             {
-                //printf ("calling DecodeAndRender\n");
-            
-                ulLastWriteTime = DecodeAndRender(ulCurrentStreamTime,
-                                                  pPkt);
-                bWroteData = TRUE;
 
-                bDone =  IsTimeGreaterOrEqual(ulLastWriteTime,ulEndTime);
+                if (pPkt->b_o_s)
+                {
+                    gotoInitialHeaderState();
+                }
+                else if (pPkt->e_o_s)
+                {
+                    bEOS = TRUE;
+                }
+                else
+                {
+                    HX_ASSERT(!bEOS);
+                }
 
+                switch (m_VorbisState) {
+                case VorbisInitialHeader:
+                    res = handleInitialHeader(pPkt);
+                    bEOS = FALSE;
+                    break;
+                    
+                case VorbisCommentHeader:
+                    if (vorbis_synthesis_headerin(&m_vi, &m_vc, pPkt) >= 0)
+                    {
+                        DPRINTF(D_VORBIS,("Got Comment Header\n"));
+                        updateTACInfo(&m_vc);
+                        m_VorbisState = VorbisCodebookHeader;
+                    }
+                    break;
+                    
+                case VorbisCodebookHeader:
+                    vorbis_synthesis_headerin(&m_vi, &m_vc, pPkt);
+                    vorbis_synthesis_init(&m_vd, &m_vi);
+                    vorbis_block_init(&m_vd, &m_vb);
+                    
+                    m_VorbisState = VorbisPlay;
+        
+                    DPRINTF(D_VORBIS,("Got Codebooks\n"));
+                    break;
+                case VorbisPlay:
+                    DPRINTF(D_VORBIS,("calling DecodeAndRender\n"));
+
+                    ulLastWriteTime = DecodeAndRender(pPkt);
+
+                    bWroteData = TRUE;
+
+                    bDone =  IsTimeGreaterOrEqual(ulLastWriteTime,ulEndTime);
+
+                    break;
+
+                default:
+                    HX_ASSERT(!"Unexpected state!");
+                    bDone = TRUE;
+                    break;
+                };
                 _ogg_free(pPkt);
             }
             else 
@@ -447,8 +538,9 @@ STDMETHODIMP CVorbisRenderer::OnDryNotification(UINT32 ulCurrentStreamTime,
             }
         }
 
-        if (!bWroteData || IsTimeLess(ulLastWriteTime, ulEndTime) &&
-            !IsRebuffering())
+        if (!bWroteData || 
+            IsTimeLess(ulLastWriteTime, ulEndTime) &&
+            !IsRebuffering() && !bEOS)
         {
             /* We ran out of data and we didn't satisfy what
              * is needed. Initiate a rebuffer so we get more
@@ -466,26 +558,42 @@ STDMETHODIMP CVorbisRenderer::OnDryNotification(UINT32 ulCurrentStreamTime,
             EndRebuffer();
         }
         
-        //printf("needed %ldms sent %ldms...\n", 
-        //       ulMinimumDurationRequired, 
-        //       (bWroteData) ? (ulLastWriteTime - ulCurrentStreamTime) : 0);
+        if (bEOS)
+        {
+            /* We don't have any more data and we've seen an
+             * end of stream packet. This could be the start
+             * of a video only stream group. Disable ODN
+             * calls for now so we don't rebuffer and
+             * transition to the VorbisInitialHeader state so 
+             * that we are ready for the next stream  and so
+             * OnPacket will process the packets. 
+             */
+            m_pCurrentStream->SetDryNotification(NULL);
+            gotoInitialHeaderState();
+        }
+
+        DPRINTF(D_VORBIS,("needed %ldms sent %ldms...\n", 
+                          ulMinimumDurationRequired, 
+                          (bWroteData) ? (ulLastWriteTime - ulCurrentStreamTime) : 0));
 
         res = HXR_OK;
     }
-        
+
+    DPRINTF(D_VORBIS,("OnDryNotification() : res %08x\n", res));
+
     return res;
 }
 
 STDMETHODIMP CVorbisRenderer::OnTimeSync(UINT32 currentPlayBackTime)
 {
-    //printf("OnTimeSync(%u)\n", currentPlayBackTime);
+    DPRINTF(D_VORBIS,("OnTimeSync(%u)\n", currentPlayBackTime));
 
     return HXR_OK;
 }
 
 STDMETHODIMP CVorbisRenderer::OnPreSeek(UINT32 timeBeforeSeek, UINT32 timeAfterSeek)
 {
-    //printf("OnPreSeek(%u, %u)\n", timeBeforeSeek, timeAfterSeek);
+    DPRINTF(D_VORBIS,("OnPreSeek(%u, %u)\n", timeBeforeSeek, timeAfterSeek));
 
     m_bInSeekMode = TRUE;
     return HXR_OK;
@@ -493,7 +601,7 @@ STDMETHODIMP CVorbisRenderer::OnPreSeek(UINT32 timeBeforeSeek, UINT32 timeAfterS
 
 STDMETHODIMP CVorbisRenderer::OnPostSeek(UINT32 timeBeforeSeek, UINT32 timeAfterSeek)
 {
-    //printf("OnPostSeek(%u, %u)\n", timeBeforeSeek, timeAfterSeek);
+    DPRINTF(D_VORBIS,("OnPostSeek(%u, %u)\n", timeBeforeSeek, timeAfterSeek));
 
     m_bInSeekMode = FALSE;
     
@@ -509,7 +617,7 @@ STDMETHODIMP CVorbisRenderer::OnPostSeek(UINT32 timeBeforeSeek, UINT32 timeAfter
 
 STDMETHODIMP CVorbisRenderer::OnPause(UINT32 timeBeforePause)
 {
-    //printf ("CVorbisRenderer::OnPause(%u)\n", timeBeforePause);
+    DPRINTF(D_VORBIS,("CVorbisRenderer::OnPause(%u)\n", timeBeforePause));
 
     return HXR_OK;
 }
@@ -521,7 +629,7 @@ STDMETHODIMP CVorbisRenderer::OnBuffering(UINT32 reason, UINT16 percentComplete)
 
 STDMETHODIMP CVorbisRenderer::OnEndofPackets(void)
 {
-    //printf ("CVorbisRenderer::OnEndofPackets()\n");
+    DPRINTF(D_VORBIS,("CVorbisRenderer::OnEndofPackets()\n"));
 
     if (m_pDepack)
     {
@@ -540,7 +648,7 @@ STDMETHODIMP CVorbisRenderer::OnEndofPackets(void)
 
 STDMETHODIMP CVorbisRenderer::EndStream(void)
 {
-    //printf ("CVorbisRenderer::EndStream()\n");
+    DPRINTF(D_VORBIS,("CVorbisRenderer::EndStream()\n"));
 
     Shutdown();
         
@@ -586,65 +694,132 @@ STDMETHODIMP CVorbisRenderer::QueryInterface(REFIID interfaceID, void **ppInterf
     return HXR_NOINTERFACE;
 }
 
-HX_RESULT CVorbisRenderer::InitAudioStream(UINT32 ulSampleRate, UINT16 usChannels)
+HX_RESULT 
+CVorbisRenderer::InitAudioStream(UINT32 ulSampleRate, UINT16 usChannels)
 {
     HX_RESULT res = HXR_OK;
 
-    //printf ("InitAudioStream(%u, %u)\n", ulSampleRate, usChannels);
+    DPRINTF(D_VORBIS,("InitAudioStream(%u, %u)\n", ulSampleRate, usChannels));
 
-    if (m_pAudioStream)
+    UINT32 uCurrentSampleRate = 0;
+    UINT32 uCurrentChannels = 0;
+
+    if (m_pCurrentStream)
     {
-        IHXAudioStream2* pStream2 = NULL;
-        
-        if (HXR_OK == m_pAudioStream->QueryInterface(IID_IHXAudioStream2, (void**)&pStream2))
-        {
-            IHXDryNotification* pDryNot = NULL;
-                                        
-            QueryInterface(IID_IHXDryNotification, (void**) &pDryNot);
-                                        
-            if (pDryNot)
-            {
-                /* Remove our OnDryNotification for this stream */
-                pStream2->RemoveDryNotification(pDryNot);
-            }
-            HX_RELEASE(pDryNot);
-        }
-        HX_RELEASE(pStream2);
+        res = m_pCurrentStream->GetStreamInfo(uCurrentSampleRate,
+                                              uCurrentChannels);
 
-        /* Release our handle on the old stream */
-        HX_RELEASE(m_pAudioStream);
+        m_pCurrentStream->SetDryNotification(NULL);
+        m_pCurrentStream = NULL;
     }
 
+    if ((uCurrentSampleRate != ulSampleRate) &&
+        m_totalSamples && uCurrentSampleRate)
+    {
+        ogg_int64_t q = m_totalSamples / uCurrentSampleRate;
+        ogg_int64_t r = m_totalSamples - q * uCurrentSampleRate;
+        m_totalSamples = 
+            q * ulSampleRate + (r * ulSampleRate) / uCurrentSampleRate;
+    }
+
+    m_audioFmt.ulSamplesPerSec = ulSampleRate;
     m_audioFmt.uChannels = usChannels;
     m_audioFmt.uBitsPerSample = 16;
-    
-    if ((m_audioFmt.ulSamplesPerSec != ulSampleRate) &&
-        m_totalSamples && m_audioFmt.ulSamplesPerSec)
-    {
-        ogg_int64_t q = m_totalSamples / m_audioFmt.ulSamplesPerSec;
-        ogg_int64_t r = m_totalSamples - q * m_audioFmt.ulSamplesPerSec;
-        m_totalSamples = q * ulSampleRate + (r * ulSampleRate) / m_audioFmt.ulSamplesPerSec;
-    }
     m_audioFmt.ulSamplesPerSec = ulSampleRate;
     m_audioFmt.uMaxBlockSize = 4096;
     
     m_ulBytesPerSec = (m_audioFmt.uChannels * 
                        (m_audioFmt.uBitsPerSample / 8) * 
                        m_audioFmt.ulSamplesPerSec);
-                                
-    res = m_pAudioPlayer->CreateAudioStream(&m_pAudioStream);
+    
+    m_pCurrentStream = findAudioStream(m_audioFmt.ulSamplesPerSec,
+                                       m_audioFmt.uChannels);
 
-    if (HXR_OK == res)
+    if (m_pCurrentStream)
     {
-        m_pAudioStream->AddDryNotification(this);
-
-        res = m_pAudioStream->Init(&m_audioFmt, m_pStreamHeaderObj);
+        m_pCurrentStream->SetDryNotification(this);
+    }
+    else
+    {
+        res = HXR_UNEXPECTED;
     }
 
     return res;
 }
 
-UINT32 CVorbisRenderer::DecodeAndRender(UINT32 ulNow, ogg_packet* pOp)
+COggAudioStreamHelper* 
+CVorbisRenderer::findAudioStream(UINT32 ulSampleRate, 
+                                 UINT16 usChannels)
+{
+    COggAudioStreamHelper* pRet = NULL;
+
+    CHXSimpleList::Iterator itr = m_audioStreams.Begin();
+
+    for (;!pRet && (itr != m_audioStreams.End()); ++itr)
+    {
+        COggAudioStreamHelper* pHlpr = 
+            (COggAudioStreamHelper*)*itr;
+        
+        UINT32 uSampRate;
+        UINT32 uChannels;
+
+        if ((HXR_OK == pHlpr->GetStreamInfo(uSampRate, uChannels)) &&
+            (uSampRate == ulSampleRate) &&
+            (usChannels == uChannels))
+        {
+            pRet = pHlpr;
+            pRet->AddRef();
+        }
+    }
+
+    if (!pRet)
+    {
+        IHXAudioStream* pAudioStream = NULL;
+        HX_RESULT res = m_pAudioPlayer->CreateAudioStream(&pAudioStream);
+
+        if (HXR_OK == res)
+        {
+            pRet = new COggAudioStreamHelper;
+
+            if (pRet)
+            {
+                pRet->AddRef(); // For return value
+                res = pRet->Init(pAudioStream);
+
+                if (HXR_OK == res)
+                {
+                    if (m_audioStreams.AddTail(pRet))
+                    {
+                        pRet->AddRef(); // For list
+                        res = pAudioStream->Init(&m_audioFmt, 
+                                                 m_pStreamHeaderObj);
+
+                        if (HXR_OK != res)
+                        {
+                            pRet->Release(); // For list
+                            m_audioStreams.RemoveTail();
+                        }
+                    }
+                    else
+                    {
+                        res = HXR_OUTOFMEMORY;
+                    }
+                }
+                            
+                if (HXR_OK != res)
+                {
+                    HX_RELEASE(pRet); // For return value
+                }
+            }
+        }
+
+        HX_RELEASE(pAudioStream);
+    }
+
+    return pRet;
+}
+
+UINT32 CVorbisRenderer::DecodeAndRender(ogg_packet* pOp)
 {
     UINT32 ulLastWriteTime = 0;
 
@@ -656,7 +831,7 @@ UINT32 CVorbisRenderer::DecodeAndRender(UINT32 ulNow, ogg_packet* pOp)
     if (vorbis_synthesis(&m_vb, pOp) == 0)
         vorbis_synthesis_blockin(&m_vd, &m_vb);
     else {
-        //printf("ERROR ? HMMMMMMM\n");
+        DPRINTF(D_VORBIS,("ERROR ? HMMMMMMM\n"));
     }
 
     UINT32 ulAudioTime = CurrentTime();
@@ -710,9 +885,10 @@ UINT32 CVorbisRenderer::DecodeAndRender(UINT32 ulNow, ogg_packet* pOp)
             {
                 ulLastWriteTime = audioData.ulAudioTime;
 
-                HX_RESULT err = m_pAudioStream->Write(&audioData);
-                //printf ("write err 0x%08x samples %d audioTime %u\n", 
-                //        err, bout, audioData.ulAudioTime);
+                HX_RESULT err = m_pCurrentStream->Write(&audioData);
+                DPRINTF(D_VORBIS,
+                        ("write err 0x%08x samples %d audioTime %u\n", 
+                         err, bout, audioData.ulAudioTime));
             }
         }
         else
@@ -745,9 +921,19 @@ void CVorbisRenderer::Shutdown()
     HX_RELEASE(m_pStream);
     HX_RELEASE(m_pPlayer);
     HX_RELEASE(m_pAudioPlayer);
-    HX_RELEASE(m_pAudioStream);
     HX_RELEASE(m_pStreamHeaderObj);
 
+    HX_RELEASE(m_pCurrentStream);
+
+    while(!m_audioStreams.IsEmpty())
+    {
+        COggAudioStreamHelper* pHlpr = 
+            (COggAudioStreamHelper*)m_audioStreams.RemoveHead();
+
+        pHlpr->Close();
+        HX_RELEASE(pHlpr);
+    }
+
     HX_DELETE(m_pDepack);
 }
 
@@ -759,7 +945,7 @@ void CVorbisRenderer::StartRebuffer()
 
     if (m_pStream && !m_bEndOfPackets)
     {
-        //printf("Starting rebuffer\n");
+        DPRINTF(D_VORBIS,("Starting rebuffer\n"));
 
         m_bRebuffering = TRUE;
         m_pStream->ReportRebufferStatus(1, 0);
@@ -772,7 +958,7 @@ void CVorbisRenderer::EndRebuffer()
 
     if (m_pStream)
     {
-        //printf("Ending rebuffer\n");
+        DPRINTF(D_VORBIS,("Ending rebuffer\n"));
         m_pStream->ReportRebufferStatus(1, 1);
     }
 }
@@ -889,3 +1075,203 @@ BOOL CVorbisRenderer::AdjustAudioData(REF(HXAudioData) audioData)
 
     return bResult;
 }
+
+void CVorbisRenderer::updateBitrateInfo(vorbis_info* vi)
+{
+    if (vi && m_pStream && m_pContext)
+    {
+        INT32 clipBitrate = 0;
+
+        if (vi->bitrate_nominal > 0)
+        {
+            clipBitrate = vi->bitrate_nominal;
+        }
+        else if (vi->bitrate_upper > 0)
+        {
+            clipBitrate = vi->bitrate_upper;
+        }
+
+        UINT32 uStreamRegID;
+        IHXRegistry* pReg = NULL;
+        if ((HXR_OK == getRegistryID(m_pStream, uStreamRegID)) &&
+            (HXR_OK == m_pContext->QueryInterface(IID_IHXRegistry,
+                                                  (void**)&pReg)))
+        {
+            CHXString propName;
+
+            if ((clipBitrate > 0) &&
+                (HXR_OK == getPropName(uStreamRegID, "ClipBandwidth",
+                                       propName)))
+            {
+                pReg->SetIntByName(propName, clipBitrate);
+            }
+        }
+        HX_RELEASE(pReg);
+    }
+}
+
+void CVorbisRenderer::updateTACInfo(vorbis_comment* vc)
+{
+    if (vc && m_pStream && m_pContext)
+    {
+        IHXStreamSource* pStreamSrc = NULL;
+        IHXRegistry* pReg = NULL;
+        UINT32 uSourceRegID;
+
+        if ((HXR_OK == m_pStream->GetSource(pStreamSrc)) &&
+            (HXR_OK == getRegistryID(pStreamSrc, uSourceRegID)) &&
+            (HXR_OK == m_pContext->QueryInterface(IID_IHXRegistry,
+                                                  (void**)&pReg)))
+        {
+            int tagInfoCount = sizeof(zm_tagInfo) / sizeof(VorbisTagMapping);
+            for (int i = 0; i < tagInfoCount; i++)
+            {
+                char* pVorbisTag = zm_tagInfo[i].m_pVorbisTag;
+                const char* pPropName = zm_tagInfo[i].m_pPropName;
+
+                int commentCount = vorbis_comment_query_count(vc, pVorbisTag);
+                
+                if (commentCount > 0)
+                {
+                    CHXString value = vorbis_comment_query(vc, pVorbisTag, 0);
+
+                    for (int j = 1; j < commentCount; j++)
+                    {
+                        value += ", ";
+                        value += vorbis_comment_query(vc, pVorbisTag, j);
+                    }
+
+                    IHXBuffer* pPropValBuf = NULL;
+                    
+                    if (HXR_OK == CreateStringBuffer(pPropValBuf, 
+                                                     value,
+                                                     m_pContext))
+                    {
+                        
+                        CHXString propName;
+
+                        if (HXR_OK == getPropName(uSourceRegID, pPropName,
+                                                  propName))
+                        {
+                            pReg->SetStrByName(propName, pPropValBuf);
+                        }
+                    }
+                    HX_RELEASE(pPropValBuf);
+                }
+            }
+        }
+        HX_RELEASE(pStreamSrc);
+        HX_RELEASE(pReg);
+    }
+}
+
+HX_RESULT 
+CVorbisRenderer::getPropName(UINT32 uBaseID, const char* pChildPropName,
+                             CHXString& propName)
+{
+    HX_RESULT res = HXR_INVALID_PARAMETER;
+
+    if (!m_pContext)
+    {
+        res = HXR_UNEXPECTED;
+    }
+    else if (pChildPropName && strlen(pChildPropName))
+    {
+        IHXRegistry* pReg = NULL;
+        IHXBuffer* pBasePropName = NULL;
+
+        res = m_pContext->QueryInterface(IID_IHXRegistry, (void**)&pReg);
+        
+        if (HXR_OK == res)
+        {
+            res = pReg->GetPropName(uBaseID, pBasePropName);
+        }
+
+        if (HXR_OK == res)
+        {
+            propName = (const char*)pBasePropName->GetBuffer();
+            propName += '.';
+            propName += pChildPropName;
+        }
+
+        HX_RELEASE(pReg);
+        HX_RELEASE(pBasePropName);
+    }
+
+    return res;
+}
+
+HX_RESULT 
+CVorbisRenderer::getRegistryID(IUnknown* pUnk, REF(UINT32) uRegID)
+{
+    HX_RESULT res = HXR_INVALID_PARAMETER;
+
+    if (pUnk)
+    {
+        IHXRegistryID* pRegID = NULL;
+
+        res = pUnk->QueryInterface(IID_IHXRegistryID, (void**)&pRegID);
+
+        if (HXR_OK == res)
+        {
+            res = pRegID->GetID(uRegID);
+        }
+
+        HX_RELEASE(pRegID);
+    }
+
+    return res;
+}
+
+void CVorbisRenderer::gotoInitialHeaderState()
+{
+    vorbis_comment_clear(&m_vc);
+    vorbis_info_clear(&m_vi);
+    
+    vorbis_info_init(&m_vi);
+    vorbis_comment_init(&m_vc);
+    
+    m_VorbisState = VorbisInitialHeader;
+}
+
+HX_RESULT 
+CVorbisRenderer::handleInitialHeader(ogg_packet* pOp)
+{
+    HX_RESULT res = HXR_FAILED;
+
+    if (vorbis_synthesis_headerin(&m_vi, &m_vc, pOp) >= 0) 
+    {
+        DPRINTF(D_VORBIS,("Got Initial Header\n"));
+        
+        updateBitrateInfo(&m_vi);
+        
+        /* check to see if the audio format is different
+         * from the default values 
+         */
+        if ((m_audioFmt.uChannels != m_vi.channels) ||
+            (m_audioFmt.ulSamplesPerSec != (UINT32)m_vi.rate))
+        {
+            /* The parameters are different. 
+             * Release our handle on the current stream
+             * and create a new audio stream object.
+             */
+            res = InitAudioStream(m_vi.rate, 
+                                  m_vi.channels);
+            
+            m_bIsGapInStreaming = TRUE;
+        }
+        else
+        {
+            m_pCurrentStream->SetDryNotification(this);
+            res = HXR_OK;
+        }
+        
+        m_VorbisState = VorbisCommentHeader;
+    } 
+    else
+    {
+        HX_ASSERT(FALSE);
+    }
+
+    return res;
+}
