@@ -1,5 +1,5 @@
 /* ***** BEGIN LICENSE BLOCK *****
- * Source last modified: $Id: prefsdialog.cpp,v 1.18.2.8.2.4 2004/09/09 18:34:32 ehayashi Exp $
+ * Source last modified: $Id: prefsdialog.cpp,v 1.18.2.10 2004/10/27 16:59:17 nhart Exp $
  * 
  * Portions Copyright (c) 1995-2004 RealNetworks, Inc. All Rights Reserved.
  * 
@@ -347,7 +347,7 @@ hpd_populate_lang_combo(GtkCombo* combo)
     
     for(i = 0; i < sizeof(g_lang_id_name_map) / sizeof(*g_lang_id_name_map); i++)
     {
-        items = g_list_insert_sorted(items, (gpointer)gettext(g_lang_id_name_map[i].lang_name), hpd_compare);
+        items = g_list_insert_sorted(items, (gpointer)gettext(g_lang_id_name_map[i].lang_name), (GCompareFunc)hpd_compare);
     }
 
     gtk_combo_set_popdown_strings(GTK_COMBO(combo), items);
